require 'curses'
require 'terminfo'

module CursorControl
  include Terminfo
  def cursor_init(stream)
    @stream = stream
    setupterm(nil,@stream)
    @bell = get_seq('bel')
    @right = get_seq('cuf1')
    @left = get_seq('cub1')
    @clr_eol = get_seq('el')
    @save_cur = get_seq('sc')
    @restore_cur = get_seq('rc')
    @to_status = getstr('tsl')
    @from_status = getstr('fsl')
    @clear = get_seq('clear')
    @to_bol = get_seq('cr')
    if @to_status.nil? or @from_status.nil? then
      @statusline = false
      @goto = get_seq('cup')
      l = ENV['LINES']
      if l.nil? then
        l = getnum('lines')
        if l == 0 then
          raise "Can't decide #lines"
        end
        @lines = l
      else
        @lines = l.to_i
      end
      c = ENV['COLUMNS']
      if c.nil? then
        c = getnum('cols')
        if c == 0 then
          c = 80
        end
      end
      @cols = c
      @change_scroll = get_seq('csr')
    else
      @statusline = true
    end
  end
  def get_seq(*capnames)
    for c in capnames
      s = getstr(c)
      return s unless s.nil?
    end
    raise "Capability "+capnames.join(" or ")+" not found"
  end
  def left(n=1)
    return if n<=0
    n.times do
      tputs(@left,@stream)
    end
  end
  def right(n=1)
    return if n<=0
    n.times do
      tputs(@right,@stream)
    end
  end
  def beep; tputs(@bell,@stream); end
  def clear; tputs(@clear,@stream); end
  def clear_to_eol; tputs(@clr_eol,@stream); end
  def to_status
    if @statusline then
      tputs(@to_status,@stream)
      return
    end
    tputs(@save_cur,@stream)
    tputs(tparm(@goto,999,0),@stream)
  end
  def from_status
    if @statusline then
      tputs(@from_status,@stream)
    else
      tputs(@restore_cur,@stream)
    end
  end
  def beginning_of_line
    tputs(@to_bol,@stream)
  end
  def erase_before(n)
    left(n)
    @stream.print ' '*n
    left(n)
  end
  def start_screen
    unless @statusline
      tputs(tparm(@change_scroll,0,@lines-2),@stream)
    end
  end
  def end_screen
    unless @statusline
      tputs(tparm(@change_scroll,0,999),@stream)
    end
  end
  def lines
    if @statusline then
      @lines
    else
      @lines-1
    end
  end
  def cols
    @cols
  end
  def message(str)
    to_status
    @stream.print str
    clear_to_eol
    from_status
    @stream.flush
  end
end

