require 'pty'
require 'cursor'
require 'skkconv'
require 'keymap'
require 'delegate'

class Fixnum
  def downcase
    if ?A <= self && self <= ?Z then
      self-?A+?a
    else
      self
    end
  end
  def upcase
    if ?a <= self && self <= ?z then
      self-?a+?A
    else
      self
    end
  end
end

class ControlledIO < SimpleDelegator
  include CursorControl
  def initialize(stream)
    if stream.kind_of?(String) then
      stream = File.open(stream,"w")
    end
    super(stream)
    cursor_init(stream)
  end
end

############### Global functions for skkfeprc description ###############
def kuten(x)
 InputContext.kuten(x)
end
def touten(x)
 InputContext.touten(x)
end

###################################### MAIN #############################
#STD_DEBUG =  File.open("/dev/pts/4","w")

unless ENV['SKKSERVER'].nil? then
  skkserver = ENV['SKKSERVER']
else
  skkserver = 'localhost'
end

skkjisyo = "~/.skk-jisyo"

####### START MAIN PROGRAM #######

### read rc file
rcfile = File.expand_path("~/.skkfeprc")
if FileTest.exist?(rcfile) then
  begin
    load(rcfile)
  rescue
  end
end

### option analysys

while ARGV[0] =~ /^-/
  case ARGV.shift
  when "-h"
    skkserver = ARGV.shift
  when "-jisyo","-jisho"
    skkjisyo = ARGV.shift
  else
    STDERR.print "
usage: rskkfep [options]
options:
  -h host          specify SKK server host
  -jisyo file      specify SKK jisyo file
"
  exit
  end
end

### start

std_screen = ControlledIO.new(STDOUT)
std_screen.start_screen
std_screen.clear

std_input = STDIN

ENV['LINES'] = std_screen.lines.to_s
ENV['COLUMNS'] = std_screen.cols.to_s
ENV['SKKFEP_PID'] = $$.to_s

system "stty -echo raw lnext ^_"
$skk = SKK.new(skkjisyo,skkserver)
std_screen.message("SKK")

r_pty,w_pty,pid = PTY.spawn("/bin/tcsh")
begin
  class << w_pty
    def erasable?
      true
    end
    def erase
      self.putc(?\b)
      self.flush
    end
    def clear
      self.flush
    end
  end

  $dispatch = SKKFEPDispatcher.new($skk,std_screen,w_pty,true)

  main = Thread.current
  Thread.new {
    GC.start
    sleep 30
  }
  begin
    wr = Thread.new {
      begin
        while true
          $dispatch.putc(std_input.getc)
        end
      rescue
        print "WR: ",$@,':',$!,"\n"
        main.raise(nil)        
      end
    }
    while true
      c = r_pty.sysread(1024)
      std_screen.write(c)
      std_screen.flush
    end
  rescue
    PTY.reset_signal
    std_screen.end_screen
    print "RD: ",$@,':',$!,"\n"
  end
end

system "stty -raw echo lnext ^V"
$skk.close

