require 'cursor'
require 'input_context'
require 'skkconv'
require 'hira2kata'

ZenkakuAlpha = [
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
	"",	"",	"",	"",	"",	"",	"",	"",
]

class SparseKeymap
  def initialize(def_item = nil)
    @default = def_item
    @map = {}
  end
  def set_default(def_item)
    @default = def_item
  end
  def [] (key)
    if @map.key?(key) then
      @map[key]
    else
      @default
    end
  end
  def []= (key,val)
    @map[key] = val
  end
  def size
    @map.size
  end
end

######## KEYMAP #######

class SKKFEPDispatcher
  NORMAL_KEYMAP = SparseKeymap.new
  KANA_KEYMAP = Array.new(128)
  CONV_KEYMAP = Array.new(128)
  ACONV_KEYMAP = SparseKeymap.new
  OKURI_KEYMAP = Array.new(128)
  SELECT_KEYMAP = Array.new(128)
  ZENKAKU_KEYMAP = SparseKeymap.new
  ENTRY_KEYMAP = SparseKeymap.new
  CODEINPUT_KEYMAP = SparseKeymap.new
  SMALL_LETTERS = "abcdefghijkmnoprstuvwxyz"
  CAPITAL_LETTERS = "ABCDEFGHIJKMNOPRSTUVWXYZ"
  CTRLS = []
  OTHERS = ' !"#$%&\'()*+,-./0123456789:;<=>?@[\\]^_`{|}~'
  VOWEL2KANA = {?a=>'',?i=>'',?u=>'',?e=>'',?o=>''}

  SKK_MODE_MSG = "SKK"
  CONVERT_MODE_MSG = "ñ"
  SELECT_MODE_MSG = "ñ"
  ZENKAKU_MODE_MSG = ""

  ######## KEYMAP CALLBACKS ########
  def through(c)
    @process.putc(c)
    @process.flush
  end
  def to_kana(c)
    set_keymap KANA_KEYMAP,@current_context.status
  end
  def to_normal(c)
    @current_context.flush_conso
    set_keymap NORMAL_KEYMAP,SKK_MODE_MSG
  end

  def start_conv(c)
    @current_context.flush_conso
    @current_context.to_hiragana
    set_keymap CONV_KEYMAP, CONVERT_MODE_MSG
    @screen.print ""
    @current_context = @conv_context.clear
    @current_context.putc(c.downcase)
  end
  def flush_without_convert(c)
    @current_context.flush_conso
    set_keymap KANA_KEYMAP,@current_context.status
    @current_context = @normal_context.clear
    str = @conv_context.output.get
    @screen.erase_before(str.size+2)
    @screen.flush
    @process.print str
    @process.flush
  end
  def flush_without_convert_and_through(c)
    flush_without_convert(c)
    through(c)
  end
  def cancel_conv_mode(c)
    @current_context.clear_conso
    @screen.erase_before(@conv_context.output.get.size+2)
    @current_context = @normal_context
    set_keymap KANA_KEYMAP,@current_context.status
  end

  def to_katakana_and_fix(c)
    @current_context.flush_conso
    set_keymap KANA_KEYMAP,@current_context.status
    @current_context = @normal_context.clear
    str = @conv_context.output.get
    @screen.erase_before(str.size+2)
    @screen.flush
    @process.print hira2kata(str)
    @process.flush
  end
  

  def kanaproc(c)
    @current_context.putc(c)
  end
  def kana_backspace(c)
    @current_context.erase
  end
  def toggle_kana(c)
    @current_context.flush_conso
    @current_context.toggle
    @screen.message(@current_context.status)  if @show_msg
  end

  # this function is not a callback
  def to_selection(yomi,okuri,extra_len)
    @candlist = @skk.candidate(yomi,okuri)
    cand = @candlist.next_cand
    if cand.nil? then
      begin_entry(@candlist.yomi)
    else
      len = yomi.size+2+extra_len;
      @screen.erase_before(len)
      @screen.print "",cand
      unless okuri.nil? then
        @screen.print okuri
      end
      set_keymap SELECT_KEYMAP, SELECT_MODE_MSG
    end
    @screen.flush
  end

  def to_okuri_vowel(c)
    c = c.downcase
    if @current_context.conso_size > 0 then
      kanaproc(c)
    else
      to_selection(@conv_context.output.get+c.chr,VOWEL2KANA[c],-1)
    end
  end
  def to_okuri_conso(c)
    c = c.downcase
    if @current_context.conso_size > 0 then
      kanaproc(c)
    else
      @current_context = @okuri_context.clear
      set_keymap OKURI_KEYMAP
      @screen.print "*"
      @current_context.putc(c)
      @okuri_conso = c
    end
  end

  def input_okuri(c)
    kanaproc(c.downcase)
  end

  def finish_okuri(c)
    kanaproc(c.downcase)
    okuri = @okuri_context.output.get
    to_selection(@conv_context.output.get+@okuri_conso.chr,okuri,okuri.size)
  end

  def okuri_backspace(c)
    e = @current_context.erase
    unless e
      # InputContext#erase == false means that there is no
      # chatacter(s) to erase.
      set_keymap CONV_KEYMAP
      @current_context = @conv_context
      @screen.erase_before(1)
    end
    @screen.flush
  end
  def cancel_okuri(c)
    @current_context.clear_conso
    @screen.erase_before(@current_context.output.get.size+1)
    set_keymap CONV_KEYMAP
    @current_context = @conv_context
    @screen.flush
  end

  def start_selection(c)
    @current_context.flush_conso
    to_selection(@current_context.output.get,nil,0)
  end

  def conv_backspace(c)
    unless @current_context.erase
      # InputContext#erase == false means that there is no
      # chatacter(s) to erase.
      set_keymap KANA_KEYMAP,@current_context.status
      @current_context = @normal_context
      @screen.erase_before(2)
    end
    @screen.flush
  end

  def fix_selection(c)
    cand = @candlist.current
    unless cand.nil?
      @candlist.fix
      fix_selection0(cand,cand.size)
    end
  end
  def fix_selection0(cand,len)
    len += 2
    if @candlist.have_okuri? then
      len += @candlist.okuri.size
    end
    @screen.erase_before(len)
    @screen.flush
    @process.print cand
    if @candlist.have_okuri? then
      @process.print @candlist.okuri
    end
    @process.flush
    sleep(0.01)
    @current_context = @normal_context.clear
    set_keymap KANA_KEYMAP, @current_context.status
    @screen.flush
  end

  def fix_selection_and_putc(c)
    fix_selection(c)
    @current_context.putc(c)
    @screen.flush
  end
  def fix_selection_and_start_conv(c)
    fix_selection(c)
    start_conv(c)
  end
  def fix_selection_and_to_normal(c)
    fix_selection(c)
    to_normal(c)
  end
  def fix_selection_and_through(c)
    fix_selection(c)
    through(c)
  end
  def fix_selection_and_toggle(c)
    fix_selection(c)
    toggle_kana(c)
  end
  def fix_selection_and_start_aconv(c)
    fix_selection(c)
    start_aconv(c)
  end

  def next_cand(c)
    cand = @candlist.current
    nextcand = @candlist.next_cand
    if nextcand.nil? then
      begin_entry(@candlist.yomi)
    else
      if @candlist.have_okuri? then
        okuri = @candlist.okuri
        @screen.erase_before(cand.size+okuri.size)
        @screen.print nextcand+okuri
      else
        @screen.erase_before(cand.size)
        @screen.print nextcand
      end
      @screen.flush
    end
  end
  def prev_cand(c)
    cand = @candlist.current
    prevcand = @candlist.prev_cand
    if prevcand.nil? then
      if @candlist.have_okuri? then
        eraselen = cand.size+@candlist.okuri.size
        @screen.erase_before(eraselen+2)
        @current_context = @conv_context
        yomi = @candlist.yomi.sub(/[a-z]$/,'')
        @screen.print "",yomi
      else
        @screen.erase_before(cand.size+2)
        @screen.print "",@candlist.yomi
      end
      if @prev_keymap == ACONV_KEYMAP then
        set_keymap ACONV_KEYMAP, CONVERT_MODE_MSG
      else
        set_keymap CONV_KEYMAP, CONVERT_MODE_MSG
      end
      @screen.flush
    else
      if @candlist.have_okuri? then
        okuri = @candlist.okuri
        @screen.erase_before(cand.size+okuri.size)
        @screen.print prevcand+okuri
      else
        @screen.erase_before(cand.size)
        @screen.print prevcand
      end
      @screen.flush
    end
  end
  def cancel_selection(c)
    cand = @candlist.current
    if @candlist.have_okuri? then
      eraselen = cand.size+@candlist.okuri.size
      @screen.erase_before(eraselen+2)
      @current_context = @conv_context
      yomi = @candlist.yomi.sub(/[a-z]$/,'')
      @screen.print "",yomi
    else
      @screen.erase_before(cand.size+2)
      @screen.print "",@candlist.yomi
    end
    if @prev_keymap == ACONV_KEYMAP then
      set_keymap ACONV_KEYMAP, CONVERT_MODE_MSG
    else
      set_keymap CONV_KEYMAP, CONVERT_MODE_MSG
    end
    @screen.flush
  end

  def start_aconv(c)
    @current_context.flush_conso
    @aconv_buffer = ''
    @screen.print ""
    set_keymap ACONV_KEYMAP, CONVERT_MODE_MSG
  end
  def aconv_putc(c)
    @screen.putc(c)
    @screen.flush
    @aconv_buffer << c.chr
  end
  def aconv_backspace(c)
    if @aconv_buffer.size > 0 then
      @screen.erase_before(1)
      @screen.flush
      @aconv_buffer[-1,1] = ''
    else
      @screen.erase_before(2)
      set_keymap KANA_KEYMAP,@current_context.status
    end
  end
  def cancel_aconv(c)
    @screen.erase_before(@aconv_buffer.size+2)
    set_keymap KANA_KEYMAP,@current_context.status
  end
  def aconv_flush_without_conv(c)
    @screen.erase_before(@aconv_buffer.size+2)
    @screen.flush
    @process.print @aconv_buffer
    @process.flush
    set_keymap KANA_KEYMAP,@current_context.status
  end
  def aconv_select(c)
    if @aconv_buffer.size == 0 then
      return
    end
    @candlist = @skk.candidate(@aconv_buffer,nil)
    cand = @candlist.next_cand
    if cand.nil? then
      @screen.beep
    else
      len = @aconv_buffer.size+2
      @screen.erase_before(len)
      @screen.print "",cand
      @screen.flush
      set_keymap SELECT_KEYMAP, SELECT_MODE_MSG
    end
  end

  def to_zenkaku(c)
    @current_context.flush_conso
    set_keymap ZENKAKU_KEYMAP, ZENKAKU_MODE_MSG
  end
  def zenkaku_putc(c)
    @process.print ZenkakuAlpha[c-32]
    @process.flush
  end

  def begin_entry(yomi)
    @doing_entry = true
    @entry_buffer.clear!
    @entry_skk = SKKFEPDispatcher.new(@skk,@screen,@entry_buffer,false,true)
    @entry_skk.set_keymap KANA_KEYMAP
    unless @in_status then
      @screen.to_status
    else
      @screen.beginning_of_line
      @screen.clear_to_eol
    end
    @screen.print yomi," "
    @screen.clear_to_eol
    @screen.flush
    set_keymap ENTRY_KEYMAP
  end
  def entry_putc(c)
    @entry_skk.putc(c)
  end
  def erase_entry(c)
    if @entry_skk.doing_entry? or @entry_skk.doing_convert? then
      @entry_skk.putc(c)
      return
    end
    @entry_buffer.erase
  end
  def finish_entry(c)
    if @entry_skk.doing_entry? then
      @entry_skk.putc(c)
      unless @entry_skk.doing_entry? then
        @screen.beginning_of_line
        @screen.clear_to_eol
        @screen.print @candlist.yomi,' ',@entry_buffer.get
        @screen.flush
      end
      return
    end
    @entry_skk.flush
    @doing_entry = false
# STD_DEBUG.print "finishing entry, yomi=",@candlist.yomi," buffer=",@entry_buffer.inspect,"\n"
    cand = @entry_buffer.get
    if cand.size == 0 then
      unless @in_status
        @screen.from_status
      end
      if @candlist.size == 0 then
        set_keymap @prev_keymap,CONVERT_MODE_MSG
      else
        set_keymap SELECT_KEYMAP,SELECT_MODE_MSG
      end
      @screen.flush
      return
    end
    @candlist.fix_newcand(cand)
    if @in_status then
      @process.print cand
      set_keymap KANA_KEYMAP
    else
      @screen.from_status
      fix_selection0(cand,@candlist.yomi.size)
    end
  end

  def start_code(c)
    @current_context.flush_conso
    @code_buffer = ''
    unless @in_status then
      @screen.to_status
      @screen.print "JIS or EUC code: "
    end
    set_keymap CODEINPUT_KEYMAP
    @screen.flush
  end
  def input_code(c)
    unless @in_status then
      @screen.putc(c)
      @screen.flush
    end
    @code_buffer << c.chr
  end
  def erase_code(c)
    if @code_buffer.size > 0 then
      unless @in_status then
        @screen.erase_before(1)
        @screen.flush
      end
      @code_buffer[-1,1] = ''
    end
  end
  def cancel_code(c)
    unless @in_status then
      @screen.from_status
      @screen.flush
    end
    @process.print "\\",@code_buffer,c.chr
    @process.flush
    set_keymap NORMAL_KEYMAP, SKK_MODE_MSG
  end
  def accept_code(c)
    if @code_buffer.size != 4 then
      @screen.beep
      cancel_code(c)
      return
    end
    unless @in_status then
      @screen.from_status
      @screen.flush
    end
    c1 = @code_buffer[0,2].hex
    c2 = @code_buffer[2,2].hex
    if c1 > 0x80 and c2 > 0x80 then
      # EUC code
      @process.putc c1
      @process.putc c2
    else
      @process.putc c1+0x80
      @process.putc c2+0x80
    end
    @process.flush
    set_keymap KANA_KEYMAP, @current_context.status
  end
  
  ###### KEYMAP INITIALIZATION ######
  def initialize(skk,screen_output,process,show_msg,in_status = false)
    @skk = skk
    @keymap = NORMAL_KEYMAP
    @prev_keymap = @keymap
    @screen = screen_output
    @process = process
    if NORMAL_KEYMAP.size == 0 then
      init_keymap
    end
    @normal_context = InputContext.new(@screen,@process)

    @conv_buffer = PseudoIOBuffer.new(@screen)
    @conv_context = InputContext.new(@screen,@conv_buffer)
    @okuri_buffer = PseudoIOBuffer.new(@screen)
    @okuri_context = InputContext.new(@screen,@okuri_buffer)

    @current_context = @normal_context

    @okuri_conso = 0
    @show_msg = show_msg
    @aconv_buffer = ''
    @code_buffer = ''
    @candlist = nil
    @in_status = in_status

    @entry_skk = nil
    @entry_buffer = PseudoIOBuffer.new(@screen,false)
    @doing_entry = false
  end

  def init_keymap
    for i in 0..31
      CTRLS[i] = i
    end
    CTRLS.push 127

    ####### NORMAL KEYMAP #######
    NORMAL_KEYMAP.set_default(:through)
    NORMAL_KEYMAP[?\C-j] = :to_kana

    ####### KANA KEYMAP #######
    CTRLS.each do |i|
      KANA_KEYMAP[i] = :through
    end
    CAPITAL_LETTERS.each_byte do |i|
      KANA_KEYMAP[i] = :start_conv
    end
    SMALL_LETTERS.each_byte do |i|
      KANA_KEYMAP[i] = :kanaproc
    end
    OTHERS.each_byte do |i|
      KANA_KEYMAP[i] = :kanaproc
    end
    KANA_KEYMAP[?l] = :to_normal
    KANA_KEYMAP[?q] = :toggle_kana
    KANA_KEYMAP[?/] = :start_aconv
    KANA_KEYMAP[?L] = :to_zenkaku
    KANA_KEYMAP[?\b]= :kana_backspace
    KANA_KEYMAP[?\\]= :start_code

    ####### CONV KEYMAP #######
    "AEIOU".each_byte do |i|
      CONV_KEYMAP[i] = :to_okuri_vowel
    end
    "BCDFGHJKMNPRSTVWXYZ".each_byte do |i|
      CONV_KEYMAP[i] = :to_okuri_conso
    end
    SMALL_LETTERS.each_byte do |i|
      CONV_KEYMAP[i] = :kanaproc
    end
    OTHERS.each_byte do |i|
      CONV_KEYMAP[i] = :kanaproc
    end
    CTRLS.each do |i|
      CONV_KEYMAP[i] = :flush_without_convert_and_through
    end
    CONV_KEYMAP[?\ ] = :start_selection
    CONV_KEYMAP[?\b] = :conv_backspace
    CONV_KEYMAP[?\C-j] = :flush_without_convert
    CONV_KEYMAP[?q] = :to_katakana_and_fix
    CONV_KEYMAP[?\C-g] = :cancel_conv_mode

    ####### OKURI KEYMAP #######
    "AEIOUaeiou".each_byte do |i|
      OKURI_KEYMAP[i] = :finish_okuri
    end
    "BCDFGHJKMNPRSTVWXYZbcdfghjkmnprstvwxyz".each_byte do |i|
      OKURI_KEYMAP[i] = :input_okuri
    end
    OKURI_KEYMAP[?\b] = :okuri_backspace
    OKURI_KEYMAP[?\C-g] = :cancel_okuri

    ####### ACONV KEYMAP #######
    ACONV_KEYMAP.set_default(:aconv_putc)
    CTRLS.each do |i|
      ACONV_KEYMAP[i] = :aconv_flush_without_conv
    end
    ACONV_KEYMAP[?\ ] = :aconv_select
    ACONV_KEYMAP[?\b] = :aconv_backspace
    ACONV_KEYMAP[?\C-g] = :cancel_aconv
    ####### SELECT KEYMAP #######

    CTRLS.each do |i|
      SELECT_KEYMAP[i] = :fix_selection_and_through
    end
    CAPITAL_LETTERS.each_byte do |i|
      SELECT_KEYMAP[i] = :fix_selection_and_start_conv
    end
    SMALL_LETTERS.each_byte do |i|
      SELECT_KEYMAP[i] = :fix_selection_and_putc
    end
    OTHERS.each_byte do |i|
      SELECT_KEYMAP[i] = :fix_selection_and_putc
    end
    SELECT_KEYMAP[?\C-j] = :fix_selection
    SELECT_KEYMAP[?\ ] = :next_cand
    SELECT_KEYMAP[?x] = :prev_cand
    SELECT_KEYMAP[?l] = :fix_selection_and_to_normal
    SELECT_KEYMAP[?q] = :fix_selection_and_toggle
    SELECT_KEYMAP[?/] = :fix_selection_and_start_aconv
    SELECT_KEYMAP[?\C-g] = :cancel_selection

    ####### ZENKAKU KEYMAP #######
    ZENKAKU_KEYMAP.set_default(:zenkaku_putc)
    CTRLS.each do |i|
      ZENKAKU_KEYMAP[i] = :through
    end
    ZENKAKU_KEYMAP[?\C-j] = :to_kana

    ####### ENTRY KEYMAP #######
    ENTRY_KEYMAP.set_default(:entry_putc)
    ENTRY_KEYMAP[?\r] = :finish_entry
    ENTRY_KEYMAP[?\b] = :erase_entry

    ####### CODEINPUT_KEYMAP #######
    CODEINPUT_KEYMAP.set_default(:cancel_code)
    "0123456789ABCDEFabcdef".each_byte do |i|
      CODEINPUT_KEYMAP[i] = :input_code
    end
    CODEINPUT_KEYMAP[?\r] = :accept_code
    CODEINPUT_KEYMAP[?\b] = :erase_code

    user_keymap_init
  end

#  utilities
  def set_keymap(keymap,mesg = false)
    @prev_keymap = @keymap
    @keymap = keymap
    if mesg and @show_msg then
      @screen.message mesg
      @screen.flush
    end
  end

  def current_keymap
    @current_keymap
  end

  def user_keymap_init
  end

  def putc(c)
    unless @keymap[c].nil? then
#STD_DEBUG.print what_keymap(@keymap),' ',@doing_entry,' ',c.chr.inspect,' => ',@keymap[c].id2name,"\n"
      self.send(@keymap[c],c)
    end
  end

  def flush
    if @keymap == KANA_KEYMAP then
      @current_context.flush_conso
    elsif @keymap == CONV_KEYMAP then
      flush_without_convert(0)
    elsif @keymao == ACONV_KEYMAP then
      aconv_flush_without_convert
    elsif @keymap == SELECT_KEYMAP then
      fix_selection(0)
    end
  end

  def doing_entry?
    @doing_entry
  end

  def doing_convert?
    @keymap == CONV_KEYMAP or @keymap == SELECT_KEYMAP or
    @keymap == ACONV_KEYMAP or @keymap == OKURI_KEYMAP or
    @current_context.conso_size > 0
  end

  # debug routines
  def what_keymap(keymap)
    if keymap == NORMAL_KEYMAP then "NORMAL_KEYMAP"
    elsif keymap == KANA_KEYMAP then "KANA_KEYMAP"
    elsif keymap == CONV_KEYMAP then "CONV_KEYMAP"
    elsif keymap == ACONV_KEYMAP then "ACONV_KEYMAP"
    elsif keymap == OKURI_KEYMAP then "OKURI_KEYMAP"
    elsif keymap == SELECT_KEYMAP then "SELECT_KEYMAP"
    elsif keymap == ZENKAKU_KEYMAP then "ZENKAKU_KEYMAP"
    elsif keymap == ENTRY_KEYMAP then "ENTRY_KEYMAP"
    elsif keymap == CODEINPUT_KEYMAP then "CODEINPUT_KEYMAP"
    else "UNKNOWN_KEYMAP"
    end
  end
end
