#
# SKK kana-to-kanji conversion
#
# by A. Ito
#
require 'socket'

class SKKConnection < TCPsocket
  def get_cand(yomi)
    syswrite "1"+yomi+" \n"
    rep = sysread(4096)
    if rep[0] == ?1 then
      # succeeded
      return rep.sub(/^1\//,'').chop.split(/\//)
    elsif rep[0] == ?4 then
      return []
    else
      raise "curious response from SKK server: #{rep}"
    end
  end
  def close
    print "0\n"
    super
  end
end

class SKKCandList
  def initialize(list,okuri)
    if okuri then
      @cand = []
      @ocand = {}
      while list.size > 1
        if list =~ /^\/\[/ then
          unless list.sub!(/\/\[([^\]]+)\]/,'')
            raise "Malformed dictionary!"
          end
          o_list = $1.split(/\//)
          o_yomi = o_list.shift
          @ocand[o_yomi] = o_list
        else
          unless list.sub!(/\/([^\/]+)/,'')
            raise "Malformed dictionary!"
          end
          @cand.push $1
        end
      end
    else
      @cand = list.sub(/^\//,'').split(/\//)
      @ocand = nil
    end
  end
  def cand
    @cand
  end
  def size
    @cand.size
  end
  def ocand
    @ocand
  end
  def okuri_cand(okuri)
    if @ocand.key?(okuri) then
      @ocand[okuri]
    else
      x = []
      @ocand[okuri] = x
      x
    end
  end
  def to_s
    s = "/"
    for c in @cand
      s << c+"/"
    end
    unless @ocand.nil?
      for okuri,cl in @ocand
        next if cl.size == 0
        s << "["+okuri+"/"
        for c in cl
          s << c+"/"
        end
        s << "]/"
      end
    end
    s
  end
  def merge(cand)
    for c in cand.cand
      unless @cand.include?(c)
        @cand.push(c)
      end
    end
    if cand.ocand.nil? then
      return self
    end
    for y,cl in cand.ocand
      if @ocand.key?(y)
        oc = @ocand[y]
        for c in cl
          unless oc.include?(c)
            oc.push(c)
          end
        end
      else
        @ocand[y] = cl
      end
    end
    self
  end
end

class SKKCandidate
  FROM_NONE = 0
  FROM_CAND = 1
  FROM_NET = 2
  FROM_OKURI = 3
  def initialize(yomi,okuri,clist,srv,skk)
    @yomi = yomi
    @okuri = okuri
    @srv = srv
    @orgcand = clist.cand
    if okuri.nil? then
      @okuricand = nil
    else
      @okuricand = clist.okuri_cand(okuri)
    end
    @skk = skk
    @clist = clist
    unless okuri.nil? then
      @cand = @okuricand
      @okuri_limit = @cand.size
      for c in @orgcand
        unless @okuricand.include?(c) then
          @cand.push(c)
        end
      end
    else
      @cand = @orgcand
      @okuri_limit = -1
    end
    @net_limit = @cand.size
    @net_cand = nil
    @n = -1
  end
  def current
    if @n < 0 then
      nil
    elsif @n < @net_limit then
      @cand[@n]
    else
      @net_cand[@n]
    end
  end
  def next_cand
    if @n < @cand.size-1 then
      @n += 1
      return @cand[@n]
    end
    return nil if @srv.nil?
    if @net_cand.nil? then
      @net_cand = @cand.clone
      for c in @srv.get_cand(@yomi)
        unless @cand.include?(c) then
          @net_cand.push(c)
        end
      end
    end
    if @n < @net_cand.size-1 then
      @n += 1
      return @net_cand[@n]
    end
    return nil
  end
  def prev_cand
    if @n > 0 then
      @n -= 1
      if @n >= @net_limit then
        @net_cand[@n]
      else
        @cand[@n]
      end
    else
      nil
    end
  end
  def fix_cand(cand,x)
    cand.each_index do |i|
      if cand[i] == x then
        cand[i,1] = []
        break
      end
    end
    cand.unshift(x)
  end
  private :fix_cand
  def fix
    return if @n < 0
    if @okuri.nil? then
      if @n < @net_limit then
        x = @cand[@n]
        fix_cand(@orgcand,x)
      else
        x = @net_cand[@n]
        @orgcand.unshift(x)
      end
    else
      if @n < @net_limit then
        x = @cand[@n]
        fix_cand(@okuricand,x)
        fix_cand(@orgcand,x)
      else
        x = @net_cand[@n]
        @okuricand.unshift(x)
        @orgcand.unshift(x)
      end
    end
    @skk.update_entry(@yomi,@okuri,@clist.to_s)
  end
  def fix_newcand(x)
    @orgcand.unshift(x)
    unless @okuri.nil?
      @okuricand.unshift(x)
    end
    @skk.update_entry(@yomi,@okuri,@clist.to_s)
  end
  def yomi
    @yomi
  end
  def okuri
    @okuri
  end
  def have_okuri?
    !@okuri.nil?
  end
  def size
    if @net_cand.nil? then
      @cand.size
    else
      @net_cand.size
    end
  end
end

class SKK
  SKK_PORT = 1178
  def initialize(dic="~/.skk-jisyo",host=nil)
    # read jisyo
    @host = host
    @dic_N = {}
    @dic_O = {}
    @server_sock = nil
    dic = File.expand_path(dic)
    @dicfile = dic
    f = File.open(dic,"r")
    @mtime = File.stat(dic).mtime
    okuri = false
    while f.gets
      if /^;; okuri-ari entries./ then
        okuri = true
        next
      elsif /^;; okuri-nasi entries./ then
        okuri = false
        next
      elsif /^;/ then
        next
      end
      yomi,list = chop.split(/ +/,2)
      if okuri then
        @dic_O[yomi] = list
      else
        @dic_N[yomi] = list
      end
    end
    return if host.nil?
    begin
      @server_sock = SKKConnection.new(host,SKK_PORT)
    rescue
      raise "Can't connect server"
    end
  end
  def server_host
    @host
  end
  def candidate(yomi,okuri)
    if okuri.nil? then
      clist = @dic_N[yomi]
      case clist
      when nil then
        clist = SKKCandList.new('/',true)
      when String then
        clist = SKKCandList.new(clist,false)
      end
	SKKCandidate.new(yomi,okuri,clist,@server_sock,self)
    else
      clist = @dic_O[yomi]
      case clist
      when nil
        clist = SKKCandList.new('/',true)
      when String then
        clist = SKKCandList.new(clist,true)
      end
      SKKCandidate.new(yomi,okuri,clist,@server_sock,self)
    end
  end
  def update_entry(yomi,okuri,entry)
    if okuri.nil? then
      @dic_N[yomi] = entry
    else
      @dic_O[yomi] = entry
    end
  end
  def update_dic
    if File.stat(@dicfile).mtime > @mtime then
      merge(@dicfile)
    end
    File.rename(@dicfile,@dicfile+"~")
    f = File.open(@dicfile,"w")
    f.print ";; okuri-ari entries.\n"
    for yomi in @dic_O.keys.sort
      clist = @dic_O[yomi]
      f.print yomi,' ',clist,"\n" if clist.size > 0
    end
    f.print ";; okuri-nasi entries.\n"
    for yomi in @dic_N.keys.sort
      clist = @dic_N[yomi]
      f.print yomi,' ',clist,"\n"  if clist.size > 0
    end
    f.close
  end
  def close
    unless @server_sock.nil?
      @server_sock.close
    end
    update_dic
  end
  def merge(dic)
    f = File.open(File.expand_path(dic),"r")
    okuri = false
    while f.gets
      if /^;; okuri-ari entries./ then
        okuri = true
        next
      elsif /^;; okuri-nasi entries./ then
        okuri = false
        next
      elsif /^;/ then
        next
      end
      yomi,list = chop.split(/ +/,2)
      cl = SKKCandList.new(list,okuri)
      if okuri then
        d = @dic_O
      else
        d = @dic_N
      end
      if d.key?(yomi) then
        if d[yomi].kind_of?(String) then
          d[yomi] = SKKCandList.new(d[yomi],okuri).merge(cl)
        else
          d[yomi].merge(cl)
        end
      else
        d[yomi] = cl
      end
    end
  end
end

